;       Program:    Music.asm - a little background music.
;       Author:     Jay Parsons, CIS 70160,340.  ATBBS "Jparsons"
;
;               Version 1.3       May 14, 1992
;       Version 1.2 was published June 25, 1990.  This revision fixes
;       one very obscure bug in the interrupt handler routine:
;               the direction flag is cleared before using string
;       instructions to restore the interrupt vector, since the state of
;       the flag is unknown when the interrupt occurs.
;               This revision does nothing to correct another problem--
;       this program cannot deinstall itself from memory completely.
;       If the timer interrupt vector has been changed by another program
;       after this one has been loaded, this program's attempt to restore
;       it will eventually cause a crash.  Such a crash is inevitable
;       anyway if this is being used as a dBASE .BIN and the timer vector
;       is reset, as leaving it alone would leave it pointing to nothing
;       when dBASE was quit or the .BIN released.  If nothing else has
;       redirected the timer interrupt, this program's efforts to uninstall
;       itself will work for most purposes, but there may be small areas
;       of memory not freed up that will cause problems if the program is
;       installed as a .COM and uninstalled several times.

;       See "Music.doc" for syntax, precautions, etc.

;       NOTE: This program cannot use a .dbf file written for Version 1.0
;       unless the order of its first two fields is reversed.

;----------------------------------------------------------------------

false           equ 0                           ; definitions required
true            equ -1                          ; for options

;                       Assembly Options

;       Change the false in the following line to true before assembly if you
;       will be debugging the program within dBASE by starting dBASE with the
;       DOS command "DEBUG dBASE."

debugging       equ false

;       Change the false in the following line to true before assembly
;       to link into a runnable .exe file convertible into a .com file.
;       Must remain false to create a file convertible to a dBASE .bin file.

DOSexe          equ false
                                                      
;----------------------------------------------------------------------

;               Program definitions

;       Do NOT change any of the following definitions except the first
;       three.

;       "maxnotes" must not exceed 7,680 for use as a dBASE .bin, since
;       the program and music would require more than the 32,000-byte
;       maximum for a .bin ( to 1.5).  Each note lengthens the .bin, and
;       memory used, by 4 bytes.  This does not affect the .com version.

;       "deftempo" may be changed however you wish.

;       "speed" may be changed, but at your peril.

maxnotes        equ 7680        ; maximum notes (.bin file only)
deftempo        equ 72          ; default tempo
speed           equ 4           ; timer tick-rate multiplier
tickdiv         equ 256/speed   ; expressed as high byte of divisor
ticksmin        equ 1152*speed  ; ticks in one minute

if maxnotes GT 7680             ; 7,680 is (32,000 - code length)/4
bufflen         equ 7680*4
else
bufflen         equ maxnotes*4
endif

slash           equ '/'
dot             equ '.'
CR              equ 0Dh
LF              equ 0Ah
deleted         equ '*'
endfile         equ 1Ah

maxbin          equ 32000       ; bytes of longest .bin allowed
envblock        equ 2ch         ; offset in PSP of environment block address

timefreqhi      equ 12h         ; clock frequency, 1,193,180, /65536
timefreqlo      equ 34dch       ; same mod 65536

readyport       equ 43h         ; timer port for signal
timer0port      equ 40h         ; timer port for channel 0, main timer
timer2port      equ 42h         ; for channel 2, speaker

speakerport     equ 61h         ; speaker control port
speakerbits     equ 00000011b   ; bits 0 and 1 control speaker

;       structure of the timer-control byte
;               bits 7 and 6 specify channel, range 0 - 2
;               bit 5 is 1 if msb of count word is to be sent
;               bit 4 is 1 if lsb ditto (?)
;               bits 3-1 are mode--3 for square wave generation
;               bit 0 is 1 if count is bcd, 0 if count is binary

spkrcount       equ 10110110b   ; speaker, both bytes, square wave, binary
clockcnt        equ 00110110b   ; main timer, both bytes, square wave, binary

timervect       equ 8           ; the hardware timer interrupt number
pic_ctl         equ 20h         ; control port for 8259A interrupt controller
EOI             equ 20h         ; end-of-interrupt code for 8259A

;----------------------------------------------------------------------

;       .dbf main header (preceding the field descriptor array) gives the
;       lengths of the header and records.

dbf_header      STRUC

memo_ind        db ?            ; .dbf version and .dbt flag
year            db ?            ; date of last update
month           db ?            ;       ditto
day             db ?            ;       ditto
rec_count       dw 2 dup(?)     ; reccount() as doubleword
header_len      dw ?            ; length of total header
rec_len         dw ?            ; length of a data record
resrvd1         dw ?
incomplete      db ?            ; dBASE IV only
encryptflag     db ?            ; ditto
resrvdlan       db 12 dup(?)    ; ditto, for LAN
mdxflag         db ?            ; also dBASE IV
resrvd2         db 3 dup(?)
dbf_header      ENDS

fdesclen        equ 20h         ; length of main header or field descriptor
lenbyte         equ 10h         ; offset in descriptor of field length byte

twofields       equ 3*fdesclen  ; .dbf main header and 2 field descriptors
minheader       equ twofields+1 ; .dbf header size if two fields only
firstlen        equ fdesclen+lenbyte ; offset in header of first field length

;       MS-DOS memory arena header (undocumented--DOS versions 2-4.01 only)
;       Headers occupy the paragraph just before each block of memory, such
;       as an environment block, a program's PSP, or a block allocated by
;       DOS to the program, such as the musicseg block here.

;       Following structure is for information, not used in the program.

arenahdr        STRUC
lastone         db ?            ; 'Z' if last in chain, otherwise 'M'
owner           dw ?            ; 0 if free, else PSP of owning program
blocklen        dw ?            ; in paragraphs.  Does not include header.
resrvd3         db 11 dup(?)    ; unknown; 0's for blocks not PSP or env.
arenahdr        ENDS

;               macros

close           MACRO handle                    ; close a file given handle
                LOCAL skipit
                mov bx,handle
                cmp bx,3                        ; don't close stdin, etc.
                jb skipit
                mov ah,3Eh
                int 21h
skipit:
ENDM

stripsp         MACRO                           ; point ds:si past spaces
                LOCAL striploop
                dec si
striploop:      inc si
                cmp byte ptr [si],' '
                loope striploop
ENDM

;----------------------------------------------------------------------

                .model tiny
                                                                      
;                       Code and Data

codeseg         Segment 'CODE'
                assume cs:codeseg

if DOSexe
                org 100h
start:
endif
                jmp init                        ; jump to intialization code

;       Data to remain resident for use playing music

EVEN
oldvect         dd ?                            ; storage for old int8 vector
playcount       dw ?                            ; ticks needed for this note
playticks       dw ?                            ; ticks played so far
clockticks      dw ?                            ; ticks left until "real" tick
nextnote        dw ?                            ; offset address of next note
pastend         dw ?                            ; offset address of music end
musicseg        dw ?                            ; segment address of music
PSPseg          dw ?                            ; our own PSP segment

;---------------------------------------------------------------------------

;       handler - routine to process ticks, play the music, and restore all

handler         proc far
                push ax
                push bx
                push cx
                push dx
                push ds
                push es
                mov ax,cs                       ; address our segment
                mov ds,ax

                assume ds:codeseg

                mov ax,playticks                ; get ticks played this note
                cmp playcount,ax                ; done with note?
                jz playnext                     ; then fetch and play next
                inc ax                          ; else just bump the count
                mov playticks,ax
                jmp short nowret                ; and return

playnext:       mov bx,nextnote                 ; point to next note
                cmp bx,pastend                  ; if any
                jae playdone                    ; if done, jump
                mov es,musicseg                 ; point with es too
                mov ax,es:[bx]                  ; duration first
                mov playcount,ax                ; save it as count
                mov playticks,0                 ; and no ticks played yet
                mov cx,es:[bx+2]                ; now the divisor
                add bx,4                        ; adjust pointer for next
                mov nextnote,bx                 ; save it
                jcxz rest                       ; if zero, turn off speaker
note:           call on                         ; else turn it on
                jmp short nowret
rest:           call off

nowret:         dec clockticks                  ; one more fast tick done
                jz passtick                     ; jump if time for a real one

                sti
                mov dx,pic_ctl                  ; tell 8259A we're finished
                mov al,EOI
                out dx,al
                pop es                          ; restore and return
                pop ds
                pop dx
                pop cx
                pop bx
                pop ax
                iret

passtick:       mov clockticks,speed            ; reset fast tick counter
                pop es                          ; restore registers
                pop ds
                pop dx
                pop cx
                pop bx
                pop ax
                jmp dword ptr cs:[oldvect]      ; chain to original handler

playdone:       mov ax,clockcnt
                out readyport,al
                xor ax,ax                       ; standard count is 0 (64K)
                out timer0port,al
                out timer0port,al
                push di
if DOSexe
                mov bx,musicseg                 ; point bx to music block
                dec bx                          ; less one for its header
                mov es,bx
                mov es:[1],ax                   ; release the block
                mov bx,PSPseg                   ; and our own program
                dec bx                          ; this is the header
                mov es,bx
                mov es:[1],ax                   ; release it too
endif

                mov es,ax                       ; set es to segment 0
                mov di,4*timervect              ; di to offset address
                mov ax,word ptr oldvect         ; reset vector without int
                cld                             ; added version 1.3
                stosw
                mov ax,word ptr oldvect+2
                stosw
                pop di
                jmp short rest                  ; turn off last note

handler         endp

;---------------------------------------------------------------------------

;       on - turn speaker on with tone divisor in cx
                            
On              proc near
                mov al,spkrcount                ; send get-ready signal
                out readyport,al
                mov ax,cx                       ; move the count to ax
                out timer2port,al               ; send low byte of count
                mov al,ah
                out timer2port,al               ; and high byte

                in al,speakerport               ; get control byte
                or al,speakerbits               ; turn speaker on
                out speakerport,al
                ret
On              endp

;       off - turn speaker off.
	
Off             proc near
                in al,speakerport               ; get control byte
                and al,not speakerbits          ; and turn speaker off
                out speakerport,al
                ret
Off             endp
                               
;       End of resident code

;----------------------------------------------------------------------

;       Code to load music into memory, adjust tempo, and set up
;       the fast clock and interrupt handler

transient       equ $

tempo           dw ?                            ; whole notes/minute
mus_handle      dw ?                            ; file handle for .mus
argptr          dw ?                            ; offset of argument string
memmess         db "Insufficient memory",CR,LF
readmess        db "Error reading .mus file",CR,LF
readend         equ $

even
if DOSexe
                dw 64 dup (?)
stk             equ $                           ; safe stack for the .com
endif
     
load:           mov ax,4202h                    ; find length of .mus file
                mov bx,mus_handle
                xor cx,cx
                mov dx,cx
                int 21h
                cmp dx,0
                jnz memerr                      ; never a whole segment
                mov dx,offset mus_buffer        ; offset to music buffer
                mov cl,4
                shr dx,cl                       ; expressed as paragraphs
ife DOSexe
                mov bx,maxbin                   ; 32000 bytes max
                sub bx,offset mus_buffer        ; less offset to the music
                cmp ax,bx                       ; all notes must fit
                ja memerr
                mov ax,cs                       ; get our code segment
                add dx,ax                       ; add paragraphs to buffer
                mov musicseg,dx                 ; and that's where music goes
else
                push ax                         ; save file length
                mov ax,PSPseg                   ; get PSP segment
                mov es,ax                       ; in two registers
                mov bx,cs                       ; get our code segment
                sub bx,ax                       ; calculate difference
                add bx,dx                       ; and add to code paras needed
                mov ah,4ah
                int 21h                         ; shrink our memory use
                pop bx                          ; get file length back
                jc memerr
                shr bx,cl                       ; file length in paragraphs
                inc bx                          ; plus one
                mov ah,48h
                int 21h                         ; ask DOS for a block
                jc memerr
                mov musicseg,ax                 ; and save its segment address
endif

                mov ax,4200h                    ; reset file ptr for reading
                mov bx,mus_handle
                xor cx,cx
                mov dx,cx
                int 21h
                jnc readmus                     ; jump if no error
                push ds

readerr:        pop ds
                mov dx,offset readmess
                mov cx,readend - offset readmess
                jmp erret

memerr:         mov dx,offset memmess
                mov cx,offset readmess - offset memmess
                jmp erret

readmus:        mov ax,3f00h                    ; read
                mov bx,mus_handle               ; the music file
                mov cx,-1                       ; all of it
                xor dx,dx                       ; into offset zero
                mov nextnote,dx                 ; which is first note
                mov playticks,dx                ; no ticks yet
                mov playcount,dx                ; and no count
                push ds
                mov ds,musicseg                 ; read into assigned segment

                assume ds:nothing

                int 21h
                jc readerr
                mov cs:pastend,ax               ; save end of music

;    adjust durations loaded for tempo specified for this playing

                mov bx,dx                       ; point to first duration
                mov si,ax                       ; and past the last
                mov cx,ticksmin                 ; get multiplier
adjloop:        cmp bx,si
                jae sethandle
                mov ax,[bx]
                mul cx
                div word ptr tempo              ; ticks for this note
                mov [bx],ax
                add bx,4                        ; on to the next
                jmp short adjloop

sethandle:      pop ds                          ; point back to our segment

                assume ds:codeseg

                mov clockticks,speed            ; fast ticks to count
                mov al,clockcnt
                out readyport,al                ; count is coming
                mov al,tickdiv/256              ; msb of new tick count
                out timer0port,al
                mov ax,tickdiv                  ; lsb
                out timer0port,al
                mov ax,3500h+timervect          ; get int 8 vector into es:bx
                int 21h
                mov word ptr oldvect,bx         ; and save it
                mov word ptr oldvect+2,es
                mov dx,offset handler           ; ds is already set
                mov ax,2500h+timervect          ; set vector to our handler
		int 21h
                xor al,al                       ; 0 signals success

ife DOSexe
endmus:         CLOSE mus_handle
                retf                            ; far return to dBASE

erret:          mov di,argptr                   ; point to argument
                mov es,PSPseg
                mov si,dx                       ; and to message
                rep movsb                       ; copy message to argument
                jmp short endmus
else
endmus:         push ax
                CLOSE mus_handle
                pop ax
                mov cx,cs:PSPseg
                mov ds,cx                       ; point ds to PSP

                assume ds:nothing

                cmp al,0                        ; error ?
                jz TSR                          ; if not, jump
                mov ah,4ch                      ; if error, keep al code
                int 21h                         ; terminate normally
TSR:
                mov bx,envblock
                mov ax,[bx]                     ; environment block address
                mov es,ax
                mov ah,49h                      ; release the environment block
                int 21h
                mov ax,cs                       ; get our code segment
                sub ax,cx                       ; calculate difference from PSP
                mov dx,offset cs:transient      ; offset to transient code
                mov cx,4
                shr dx,cl                       ; divide by 16
                inc dx                          ; one more won't hurt
                add dx,ax                       ; add to segment difference
                mov ax,3100h                    ; terminate but stay resident
                int 21h                         ; back to DOS with 0

erret:          mov bx,2                        ; standard error file
                mov ax,4001h                    ; al=1 for return to DOS
                int 21h                         ; print message
                jmp short endmus
align 16                                        ; this is not necessarily
mus_buffer      equ $                           ; where music will be loaded,
endif                                           ; just end of memory in use

;---------------------------------------------------------------------------

;       Start of program and code to initialize, compile, read files, etc.

;       Initialization variables

openerrmess     db "Can't find, or can't open, music .dbf file.",CR,LF
dbf_extension   db '.DBF',0
mus_extension   db '.MUS',0

durlen          dw ?                            ; length of the duration field
notelen         dw ?                            ; length of note/divisor field
dbf_handle      dw ?                            ; their DOS handles
compiled        dw ?
mus_stamp       dw 2 dup (?)

switch          db ?
mus_dur         dw ?                            ; write buffer for .mus file
mus_note        dw ?

;       Execution begins here

init:
if              debugging
                int 3                           ; pass control to DEBUG
endif

if DOSexe
;       This routine adjusts things so DOS command-line argument looks like
;       a dBASE argument, allowing use of .bin code to decode it.

                mov bx,80h                      ; argument portion of PSP
                mov al,byte ptr [bx]            ; command count
                cmp al,0                        ; if any
                jnz init1
                mov ax,4c01h                    ; no argument is an error
                int 21h
init1:          cbw                             ; command count as a word
                mov si,ax
                inc bx                          ; point bx to argument
                mov byte ptr ds:[bx+si],0       ; replace 0dh with ASCIIZ
                mov sp,offset stk               ; protect stack
endif

;       Music program proper starts here
                            
                mov si,bx                       ; point ds:si to argument
                mov cs:argptr,bx                ; save it
                mov cs:PSPseg,ds                ; and its segment
                mov cx,-1                       ; can't go this far
                STRIPSP                         ; remove spaces
                mov di,si
                mov ax,ds
                mov es,ax                       ; point es:di to it
                xor al,al                       ; look for the ASCIIZ
                mov cx,-1
                repne scasb
                not cx                          ; and convert the count
                mov ax,cs                       ; destination is our segment
                mov es,ax

                mov di,offset es:dbf_name       ; copy the argument
                rep movsb                       ; including the ASCIIZ
                mov ds,ax                       ; no more need for argument

                assume ds:codeseg,es:codeseg    ; nor for other segments

                mov switch,false
                mov tempo,false
                mov dbf_handle,0                ; set flags for no files open
                mov mus_handle,0
                call parse                      ; parse argument
                cmp tempo,false
                jnz fixtempo
                mov tempo,deftempo
fixtempo:       mov ax,tempo                    ; multiply tempo by 64, the
                xor dx,dx                       ; length of a whole note, for
                mov cx,64                       ; fraction of a minute per unit
                mul cx                          ; of duration
                mov tempo,ax
                call open                       ; open the file(s)
                jc initerr
                                                                           
compile:        cmp compiled,true               ; is .mus file ok?
                jnz readhead
                jmp load                        ; if it is, just load it
readhead:       call read_header                ; else read .dbf header
                jc initerr
                call readwrite                  ; make the .mus file
                jc initerr
                jmp load                        ; and jump back to load it

initerr:        mov dx,offset openerrmess
                mov cx,offset dbf_extension - offset openerrmess
                jmp erret

;       End of main procedure

;---------------------------------------------------------------------------

;                       Subroutines

;       parse - parse argument for filename, tempo and switch.
;       Capitalizes filename at dbf_fname.  Sets switch, tempo.

parse           proc near
                mov si,offset dbf_name
                xor dx,dx                       ; two falses
                mov bx,dx                       ; third
                mov di,dx                       ; fourth
                mov ah,dh                       ; fifth

caploop:        lodsb
                cmp al,0
                jnz getchar0
                cmp di,0
                jnz capdone
                mov di,si
                dec di
capdone:        jmp movename

getchar0:       cmp al,'0'                      ; a digit?
                jb getchar1
                cmp al,'9'
                ja getchar1
                cmp dh,true                     ; already got tempo?
                jz getchar1                     ; then nothing special
                cmp bx,0                        ; have we started filename?
                jz gettempo                     ; if not, should be tempo
                cmp di,0                        ; unless we're in the name
                jz getchar1

gettempo:       dec si                          ; back up
                push ax
                push bx
                mov cx,5                        ; 5 digits max for a tempo
                call atob                       ; convert
                mov tempo,ax
                pop bx
                pop ax
                mov dh,true                     ; set flag
                jmp short caploop

getchar1:       cmp al,slash                    ; signal for a switch?
                jnz getchar2
                mov al,[si]
                and al,5fh                      ; capitalize
                cmp al,'D'                      ; divisors?
                jnz getchar2
                cmp byte ptr [si+1],'0'         ; followed by alphanumeric?
                jae getchar2                    ; if so, no switch
                cmp dl,true                     ; nor if already gotten
                jz getchar2
                mov switch,al                   ; set the switch
                mov dl,true                     ; set the other flag
                inc si                          ; point past it
                jmp short caploop

getchar2:       cmp al,'0'                      ; alphanumeric?
                jb getchar4                     ; if not, continue
                cmp bx,0                        ; already have address?
                jnz getchar3
                mov bx,si                       ; save address of name
                dec bx

getchar3:       cmp al,'a'
                jb caploop
                and byte ptr [si-1],5fh         ; capitalize it
                jmp short caploop

getchar4:       cmp al,dot                      ; a dot in name?
                jnz getchar5
                cmp ah,0                        ; only want one
                jnz getchar5
                cmp bx,0                        ; and must be in name
                jz getchar5
                cmp di,0
                jnz getchar5
                mov ah,al                       ; set its flag if so
                jmp caploop

getchar5:       cmp bx,0                        ; have we a possible name?
                jz getchar6                     ; if not, ignore whitespace
                cmp di,0                        ; has name already ended?
                jnz getchar6                    ; if so, ignore it too
                mov di,si
                dec di                          ; else save end address

getchar6:       jmp caploop

;       move name to front of buffer

movename:       mov si,bx
                mov cx,di
                sub cx,si
                mov di,offset dbf_name
                cmp bx,di                       ; is filename already first?
                jnz moveit
                add di,cx
                jmp short endname
moveit:         rep movsb
endname:        mov byte ptr [di],0             ; add ASCIIZ in case needed
                cmp ah,dot                      ; already got a dot?
                jz gotdot                       ; then move on
                mov si,offset dbf_extension     ; add .dbf,0
                mov cx,5
                rep movsb
     
gotdot:         std
                mov cx,6                        ; how long is extension?
                mov al,dot
                repne scasb
                cld
                inc di                          ; point di to the dot
                mov si,offset dbf_name
                mov cx,di
                sub cx,si
                mov di,offset mus_name
                rep movsb                       ; copy name for .mus file
                mov si,offset mus_extension     ; add .mus,0
                mov cx,5
                rep movsb
parsedone:      ret
parse           endp

;---------------------------------------------------------------------------

;       open - open the two files, if they exist.  Compare date/time stamps
;       and set compiled flag.

open            proc near
                mov compiled,false
                mov dx,offset mus_name
                mov al,2
                call opencheck
                jc omuserr
                mov mus_handle,bx
                mov mus_stamp,cx
                mov mus_stamp+2,dx
                mov al,0
                mov dx,offset dbf_name
                call opencheck
                jc fopenerr
                mov dbf_handle,bx
                cmp dx,mus_stamp+2
                ja recompile
                jb noneed
                cmp cx,mus_stamp
                ja recompile
noneed:         mov compiled,true
                clc
recompile:      ret

omuserr:        cmp ax,2                        ; file not found
                jnz fopenerr
                mov al,0
                mov dx,offset dbf_name
                call opencheck
                jc fopenerr
                mov dbf_handle,bx
                mov ax,3C00h
                xor cx,cx
                mov dx,offset mus_name
                int 21h
                jc fopenerr
                mov mus_handle,ax
                jmp short recompile

fopenerr:       stc
                ret
open            endp

;---------------------------------------------------------------------

;       open a file named at ds:dx for mode in al and get time/date
;       stamp.  Return with carry if error, else handle in bx and time
;       and date in cx and dx

opencheck       proc near
                mov ah,3Dh                      ; open for mode in al
                int 21h
                jc openerr
                mov bx,ax
                mov ax,5700h                    ; get date and time
                int 21h
openerr:        ret
opencheck       endp

;---------------------------------------------------------------------------

;       read_header - read the header of the .dbf file, save key data
;       returns with carry set in case of i/o error

read_header     proc near
                mov ax,3f00h
                mov bx,dbf_handle
                mov cx,twofields                ; read two field descriptors
                mov dx,offset dbf
                int 21h
                jc rherr
ife DOSexe
                mov ax,dbf.rec_count            ; be sure not too many records
                mov dx,dbf.rec_count+2
                mov cx,4
                mul cx
                cmp dx,0
                jnz rherr
                mov cx,maxbin                   ; maximum offset possible
                sub cx,ax                       ; less bytes of music
                cmp cx,offset mus_buffer        ; must be at or past buffer
                jb rherr
endif
                xor cx,cx
                mov dx,dbf.header_len           ; bx=handle already
                cmp dx,minheader                ; be sure at least 2 fields
                jc rherr
                mov ax,4200h                    ; point past header for read
                int 21h
                jc rherr
                mov ah,ch
                mov bx,offset dbf+firstlen
                mov al,byte ptr [bx]            ; length of field 1
                mov durlen,ax
                add bx,20h
                mov al,byte ptr [bx]            ; length of field 2
                mov notelen,ax
                clc
                ret

rherr:          stc
                ret
read_header     endp

;---------------------------------------------------------------------------

;       readwrite - read fields from the .dbf and copy binary equivalents
;       (or divisor if switch) to .mus.

readwrite       proc near

rwnext:         mov ah,3Fh
                mov bx,dbf_handle
                mov cx,dbf.rec_len
                mov dx,offset dbf_buffer
                mov si,dx
                int 21h
                jc rwerr
                cmp ax,dbf.rec_len
                jz delcheck
                cmp ax,0
                jz rwdone
                cmp byte ptr [dbf_buffer],endfile
                jz rwdone
                jmp short rwerr

delcheck:       lodsb
                cmp al,'*'
                jz rwnext                       ; ignore deleted records

;       read note; convert frequency to divisor

                mov cx,notelen
                STRIPSP
                inc cx
                call atob
                cmp switch,false
                jnz savediv
                cmp ax,19
                jae makediv
                xor ax,ax
                jmp short savediv
makediv:        mov cx,ax
                mov dx,timefreqhi       ; divide timer frequency by
                mov ax,timefreqlo       ; requested frequency
                div cx                  ; to obtain pulse count
savediv:        mov mus_note,ax

;       read duration and save

nextdur:        mov cx,durlen
                STRIPSP
                inc cx
                call atob
                mov mus_dur,ax

;       write it to file

                mov ah,40h
                mov bx,mus_handle
                mov cx,4
                mov dx,offset mus_dur
                int 21h
                jc rwerr
                jmp rwnext

rwerr:          CLOSE dbf_handle
                stc
                ret

rwdone:         CLOSE dbf_handle
                ret
readwrite       endp

;---------------------------------------------------------------------------

;       Atob - ASCII to binary conversion routine.
;       Converts CX or fewer ASCII digits at DS:SI to binary value in AX.
;	Does not check for overflow but returns value mod 65,536.
;       Treats any non-digit as terminator.
;	Cannot handle negative numbers.
;	Destroys AX,BX.  On return, SI points past terminator.

Atob            proc near
                xor bx,bx               ; initialize the number

next_digit:     lodsb                   ; get a character
                sub al,'0'              ; else make it hex
                jb atob_done            ; done if less than 0
                cmp al,9
                ja atob_done            ; or more than 9

                cbw                     ; clear ah
                sal bx,1                ; old number times 2
                add ax,bx               ; new digit plus old number*2
                sal bx,1                ; old number times 4
                sal bx,1                ; old number times 8
                add bx,ax               ; add for 10*old plus new
                loop next_digit

atob_done:      mov ax,bx
                ret
Atob            endp

;-------------------------------------------------------------------------

;       Caution - the music buffer includes the 188 bytes included
;       in dbf_name, mus_name and dbf as well as those allocated
;       in the last line.  Do not rearrange the code below heedlessly.

ife DOSexe
align 16
mus_buffer      equ $
endif

dbf_name        db 78 dup (?)           ; path & name of .dbf
mus_name        db 78 dup (?)           ; and of .mus
dbf             dbf_header <>           ; storage of the header info
dbf_buffer      equ dbf + 12

ife DOSexe
                db bufflen-188 dup(0)   ; buffer for .dbf file, music
endif
CODESEG         ends
IF DOSexe                               ; this will cause harmless
                end start               ; "open conditional" message
else                                    ; at assembly since the
                end                     ; "endif" is not reached.
endif
